#ifndef LEVEL_H
#define LEVEL_H

/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

#include <list>
using namespace std;

class Level;

#include "pad.h"
#include "ball.h"
#include "item.h"
#include "stone.h"
#include "powerup.h"

#define CLS_FIELD 2

/* Note that field_info contains information about pads so that walls are drawn correctly, but pads must still be added separately */
struct field_info {
  unsigned int id; // must be zero
  unsigned char cls; // must be CLS_FIELD
  unsigned char ev;
  int width;
  int height;
  unsigned char pads; // sum of somethingPAD (pad.h)
};

class Level {
public:
  /* the "passive" parameter prevents creation of a new ball when
     there's none (for use in client mode), until I find something better */
  Level(bool passive);
  ~Level();
  
  /* Parameter: the name of a file contained in the levels/
     directory. If not set, opens levels/(levelnum) */
  bool load(const char *lvl=0);
  //  Cell *getCell(int x,int y);

  /** Add the given object in the list of items (this is called from
      the Item constructor). If the item's id is 0 then a new one is
      created and returned. */
  unsigned int register_item(Item *i);

  /** Only used to resize the level */
  void process(struct item_info *data,size_t len);

  /* This will eventually be dealt with by register_item. */
  void add_stone(Stone *s);
  void add_pad(Pad *p);
  void add_ball(Ball *b);

  /* Remove the item from the list (this is called from the item
     destructor) */
  void unregister(Item *i);

  /* Get the item of given id, if any, or NULL otherwise */
  Item *get_item(unsigned int id);

  /* Forward the poll() request to all items in the level. This should
     be called when there's a new client */
  void poll();

  /* Pass to the next level */
  void skip();

  /** returns if the given trajectory intersects *any* object in the
      level (if so, return a double between 0 and 1 indicating where,
      sets n to the normal and v to the speed of the object on the
      path.
      
      Parameters:
      (x;y): Position of the object at time (now-t)

      (vx;vy): Speed of the object

      t: Time interval in which to search for collisions

      n: gets set to the normal vector of the obstacle in case of
         collision

      p: In case the collision was with a pad, gets set to a pointer
         to that pad

      v: gets set to the speed of the obstacle in case of collision.

      trigger: Whether to call the trigger() method of stones in the
         path (if any)

      o: The object being on that trajectory

      Return value: Between 0 (t seconds ago) and 1 (now), when a
      collision occurred, or 2 if no collision occurred.
  **/
  double check_collision(double x,double y,double vx,double vy,
			 double t,
			 double *n=0,double *v=0,Pad **p=0,
			 bool trigger=true, Littlecube *o=0);

  /* Call this before starting a collision detection sequence. It is
     used to avoid two subsequent collisions against the same object
     (which would necessarily be due to an approximation error */
  void reset_coll();

  /* Returns whether the given point is inside the level. */
  bool is_in_level(double x,double y);

  int get_width();
  int get_height();

  void render();

  /* t: how many seconds passed since the last update(). Returns false
     if the game is over. */
  bool update(double t);

  /* the parameters are basically the mouse pointer position, ranging
     from 0 to 1 in both dimensions, and (0,0) being bottom left. */
  void movePads(double x,double y);

  /* Move the pads by the given amounts */
  void shiftPads(double dx,double dy);

  /* Release every ball that is currently captured */
  void releaseBalls();

  /* Yipee! */
  void extra_ball(Pad *p);

 private:
  /* Free all resources, either because we changed level, or because
     we're closing the game */
  void clear();

  bool passive;

  /* The level currently being played */
  //  Level *level;
  int levelnum;
  bool levelDone;
  double fadeout; // if >0 then set light intensity to fadeout/FADETIME
		    // and decrement it for each passing millisecond.

  /** Kill the current ball and start with a new one (if there is one)
      Update to support many balls if it gets needed again.. */
  //void lost();
 
  int wid;
  int hei;

  /* Cells are just an optimisation to avoid doing more collision
     checking than required. Every solid object should register
     itself in every cell it touches, and reciprocally any object
     wanting to do collision detection should query every cell it
     touches. Then only objects having a cell in common are tested
     for collision.

     An object can be in more than one cell and a cell may contain
     more than one object.

     NOTE: This isn't actually implemented! :)*/
  //  Cell *cells;

  /* The union of stones, pads, balls, powerups. */
  list<Item *> items;
  unsigned int next_item; // the id the next created item will get

  //  int scount; // how many actual stones are in the "stones" array. The
	      // allocated array size is wid*hei (for now)
  list <Stone *> stones;

  /* How many stones need to be broken before the level is
     completed */
  int to_break;

  list <Pad *> pads;
  //  int padcount;
  Pad *favpad; // the latest pad involved in a triggering collision

  /* bitfield suming somethingPAD values from pad.h */
  int whichpads;

  /* If the latest collision in the current collision sequence was a
     pad then it is this one. See reset_coll() and
     check_collision() */
  Pad *lastcoll;

  list <Ball *> balls;

  /* Power up on screen. Can't have more than one at a time */
  Powerup *powerup;
};

#endif
