#include "region.h"

/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

#include <math.h>

#include "base.h"
#include "main.h"

Region::Region() {
  reset();
}

Region::~Region() {}

void Region::reset() {
  neutral = true;

  xa = 0;
  xz = 0;
  ya = 0;
  yz = 0;

  dxa = 0;
  dxz = 0;
  dya = 0;
  dyz = 0;
}

void Region::include(double x,double y) {
  if (neutral) {
    xa = x;
    xz = x;
    ya = y;
    yz = y;
    neutral = false;
  } else {
    if (x<xa) xa = x;
    if (x>xz) xz = x;
    
    if (y<ya) ya = y;
    if (y>yz) yz = y;
  }
}

void Region::include(Region *that) {
  if (that->neutral) return;
  if (neutral) {
    set_to(that);
  } else {
    this->xa = min(this->xa,that->xa);
    this->xz = max(this->xz,that->xz);
    this->ya = min(this->ya,that->ya);
    this->yz = max(this->yz,that->yz);
  }
}

void Region::set_to(Region *that) {
  reset(); // deals with the (d..).

  this->xa = that->xa;
  this->xz = that->xz;
  this->ya = that->ya;
  this->yz = that->yz;
  this->neutral = that->neutral;
}

void Region::smooth_set_to(Region *that, double t) {
  /* t assumed much smaller than 1 */
  dxa = dxa * (1-t) + t*(that->xa-this->xa)/2.5;
  dya = dya * (1-t) + t*(that->ya-this->ya)/2.5;
  dxz = dxz * (1-t) + t*(that->xz-this->xz)/2.5;
  dyz = dyz * (1-t) + t*(that->yz-this->yz)/2.5;

  this->xa += dxa * t;
  this->ya += dya * t;
  this->xz += dxz * t;
  this->yz += dyz * t;

  neutral = false;
}

double *Region::centre() {
  static double r[2];
  r[0] = (xz+xa)/2;
  r[1] = (yz+ya)/2;
  return r;
}

double Region::width() {
  return max(0.0,xz-xa);
}

double Region::height() {
  return max(0.0,yz-ya);
}

double Region::get_target_angle(bool vertical, double cam, double camz) {
/* first, see at what angles are the top and bottom of the requested
     area */
  double angz,anga,fov;
  if (vertical) {
    fov = PFOV;
    angz = atan((yz-cam)/camz);
    anga = atan((ya-cam)/camz);
  } else {
    fov = PFOV*screen_ratio;
    angz = atan((xz-cam)/camz),
    anga = atan((xa-cam)/camz);
  }
  
  // if (vertical) {
  //   printf("%f - [%f;%f] - %f",anga,-fov/2,fov/2,angz);
  // }

  if (angz>(fov/2)) { // top hidden
    if (anga>angz-fov) { // bottom still visible when looking up ..
      // printf("top align\n");
      return angz-(fov/2);  // .. to see the top
    } else { // if not, just look in the middle of the area, but won't
	     // see everything :-(
      // printf("middle\n");
      return (angz+anga)/2;
    }
  } else if (anga < -(fov/2)) { // bottom hidden
    if (angz>anga+fov) { // top still visible when looking down ..
      // printf("bottom align\n");
      return anga+(fov/2);  // .. to see the bottom
    } else { // if not, just look in the middle of the area, but won't
	     // see everything :-(
      // printf("middle\n");
      return (angz+anga)/2;
    }
  } else { // both ends visible when looking straight
    // printf("all clear\n");
    return 0;
  }
}

double Region::get_extraz(bool vertical, double cam, double camz, double ang) {
  double hfov;
   if (vertical) {
    hfov = PFOV/2;
  } else {
     hfov = (PFOV*screen_ratio) /2;
  }

  /* 1. what point do we see in the centre of the screen? */
  double m = cam+tan(ang)*camz;
  /* 2. How much do we see when one unit away from the 0-plane? */
  double ea = tan(ang)-tan(ang-hfov); // first half of screen
  double ez = tan(ang+hfov)-tan(ang); // second half
  /* 3. How far away from the 0-plane should we be to see everything? */
  double z;
  if (vertical) {
    z = max( (m-ya)/ea , (yz-m)/ez );
// #ifdef DEBUG
//     printf("%f [%f + %f + %f] %f: %f | %f - %f -> ",ya,ea,m,ez,yz,(m-ya)/ea , (yz-m)/ez,camz);
// #endif
  } else {
    z = max( (m-xa)/ea , (xz-m)/ez );
  }
  /* 4. We're already at distance camz, need (z-camz) more. Every unit
     moved backwards takes us cos(ang) units away from the xy-plane */
  return (z-camz) / cos(ang);
}


void Region::paint() {
  glBegin(GL_LINE_LOOP);
  glVertex3f(xa,ya,0.6);
  glVertex3f(xz,ya,0.6);
  glVertex3f(xz,yz,0.6);
  glVertex3f(xa,yz,0.6);
  glEnd();
}

void Region::print() {
  if (neutral)
    printf(" [neutral]\n");
  else
    printf(" [%f,%f]x[%f,%f]\n",xa,xz,ya,yz);
}
