#include "stone.h"
#include "main.h"

#include "boom.h"
#include "dynamo.h"
#include "hard.h"
#include "moving.h"
#include "shallow.h"
#include "slash.h"
#include "teleport.h"
#include "unbreakable.h"

/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

/* in how many pieces does a stone break? (Don't just blindly change
   this value but also update the trigger() method.) */
#define PIECECOUNT 4

Stone::Stone(unsigned int id) : Item(id) {
  /* So that if you render() a Stone before init()ing it you don't get
     funny results */
  this->gone = true;
  this->pieces = 0;
}
Stone::~Stone() {}

void Stone::init(char ch,int x,int y) {
  //  printf("initialising stone %p at (%d,%d)\n",this,c->getX(),c->getY());
  this->ch = ch;
  this->x = x;
  this->y = y;

  this->gone = false;

  poll();
}

void Stone::poll() {
  struct stone_info data;
  data.cls = CLS_STONE;
  data.ev  = EV_INIT;
  data.ch  = ch;
  data.x   = x;
  data.y   = y;
  push((struct item_info*)&data,sizeof(struct stone_info));  
}

void Stone::process(item_info *data,size_t len) {
  struct stone_info *info = (struct stone_info*)data;
  switch (info->ev) {
  case EV_INIT:
    this->x = info->x;
    this->y = info->y;
    this->gone = false;
    break;
  case EV_BROKE:
    kill();
    break;
  default:
    printf("Warning unknown Stone event %d received.\n",info->ev);
  }
}

double Stone::getX() {
  return x;
}

double Stone::getY() {
  return y;
}

double Stone::check_collision(double x,double y,double dx,double dy, double *n, double *v,bool triggers) {
  double shortest=2; // shortest collision found so far
  double cur; // distance of current obstacle

  if (gone) return 2;

  /* s=0: edges with smaller coordinate, s=1: edges with larger
     coordinate */
  for (int s=0;s<2;s++) {
    /* horizontal edge */
    cur = wall_collision(x,y, dx,dy, getX(),getX()+1, getY()+s);
    if (cur<shortest) {
      //      printf("H%d",s);
      shortest = cur;
      n[0]=0; n[1]=2*s-1;
      v[0]=0; v[1]=0;
    }
    /* vertical edge */
    cur = wall_collision(y,x, dy,dx, getY(),getY()+1, getX()+s);
    if (cur<shortest) {
      //      printf("V%d",s);
      shortest = cur;
      n[0]= 2*s-1; n[1]=0;
      v[0]=0; v[1]=0;
    }
  }
  return shortest;
}

bool Stone::breakable() { return true; }

bool Stone::trigger(Littlecube *b) {
  kill();
  return true;
}

void Stone::activate() {}

void Stone::kill(bool piecetriggers) {
  crack[rand()%4].play();

  struct stone_info data;
  data.cls = CLS_STONE;
  data.ev  = EV_BROKE;
  push((struct item_info*)&data,sizeof(struct stone_info));  

  gone = true;

  double x = getX();
  double y = getY();

  pieces = new Littlecube*[PIECECOUNT];
  for (int i=0;i<PIECECOUNT;i++) {
    pieces[i] = new Littlecube(0);
  }
  pieces[0]->init(x+0.25,y+0.25,0.5f,piecetriggers);
  pieces[0]->set_speed(-3,-3);
  pieces[0]->set_delta_radius(-1);

  pieces[1]->init(x+0.25,y+0.75,0.5f,piecetriggers);
  pieces[1]->set_speed(-3, 3);
  pieces[1]->set_delta_radius(-1);

  pieces[2]->init(x+0.75,y+0.25,0.5f,piecetriggers);
  pieces[2]->set_speed( 3,-3);
  pieces[2]->set_delta_radius(-1);

  pieces[3]->init(x+0.75,y+0.75,0.5f,piecetriggers);
  pieces[3]->set_speed( 3, 3);
  pieces[3]->set_delta_radius(-1);
}

void Stone::update(double t, Level *lvl) {
  if (pieces) {
    for (int i=0;i<PIECECOUNT;i++) {
      pieces[i]->update(t,lvl);
    }
    if (pieces[0]->has_vanished()) {
      for (int i=0;i<PIECECOUNT;i++) {
	delete pieces[i];
      }
      delete [] pieces;
      pieces = 0;
    }
  }
}

void Stone::render() {
  glColor3f(0,0.5f,0);
  render0();

}

bool Stone::render0() {
  if (gone) {
    if (pieces) {
      for (int i=0;i<4;i++) {
	pieces[i]->render();
      }
    }
    return false;
  }

  //    glEnable(GL_LIGHTING);
  glPushMatrix(); {
    glTranslatef(getX(),getY(),0);

    render_cube(0.95, 0.95, 0.5,true);
  } glPopMatrix();

  //  glDisable(GL_LIGHTING);
  return true;
}

Stone *new_stone(char ch,int id) {
  //      printf("#%c",c);
  switch (ch) {
  case '*':
    return new Stone(id);
  case '~':
    return new Moving(0,false,id);
  case '.':
    return new Shallow(id);
  case 'w':
    return new Moving(0,true,id);
  case '#':
    return new Unbreakable(id);
  case '/':
    return new Slash(false,id);
  case '\\':
    return new Slash(true,id);
  case 'B':
    return new Boom(id);
  case 'D':
    return new Dynamo(id);
  case 'H':
    return new Hard(id);
  case 's':
    return new Moving(1,false,id);
  case 'S':
    return new Moving(1,true,id);
  case 'T':
    return new Teleport(id);
    /* Note: you still need to do the set_target() thing */
  case ' ':
    return NULL;
  default:
    printf("Illegal stone character '%c'\n",ch);
    return NULL;
  }
}
